import { Clonable, Keypair, PK } from '..';
import { Me, Peer } from './peer';
import { Node } from './node';
import { Leaf } from './leaf';
export declare class ART implements Clonable {
    me: Me;
    root: Node;
    setupMessage: Uint8Array;
    stage: Keypair;
    private constructor();
    /**
     * Clone ART instance
     */
    clone(): ART;
    /**
     * Get my leaf
     */
    get leaf(): Leaf;
    /**
     * Get leaf of any peer
     * @param peer can be either Peer instance (then compared with ===), or peer identity PK / ephemeral PK
     */
    leafOf(peer: Peer | PK): Leaf | undefined;
    /**
     * Get first leaf, that is a leaf of tree initiator
     */
    get initiator(): Leaf;
    /**
     * Check if tree is made by initiator
     */
    get isInitiator(): boolean;
    /**
     * Get list of tree Peers
     */
    get peers(): Peer[];
    /**
     * Get all leafs
     */
    get leafs(): Leaf[];
    /**
     * Derive current stage key from the tree.
     *
     * @returns stage key
     */
    private deriveStageKey;
    /**
     * Initiate new ART Tree with given members. First member must be instance of `Me`, that is tree initiator.
     *
     * @param peers tree members
     * @returns
     */
    static initiate(peers: Peer[]): ART;
    /**
     * Reconstruction of a tree created by `ART.initiate()` from setup message
     *
     * @param me my peer
     * @param message setup message from initiator peer
     * @returns
     */
    static fromSetupMessage(me: Me, message: Uint8Array): ART;
    /**
     * Reconstruction of a tree created by `ART.split()` from split snapshot & message
     *
     * @param me my peer
     * @param snapshot last known snapshot of the tree
     * @param splitMessage split message
     * @returns
     */
    static fromSplitMessage(me: Me, snapshot: Uint8Array, splitMessage: Uint8Array): ART;
    /**
     * Update my ephemeral key with new one. This is main "ratcheting" method, it updates
     * stage key. Returned message must be distributed to all other peers, preferrably in encrypted form.
     *
     * @returns update key message
     */
    updateKey(key?: Keypair): Uint8Array;
    /**
     * Process key update message from another peer. Updates stage key locally.
     */
    processKeyUpdate(message: Uint8Array): void;
    /**
     * Split one of the leafs into 2 adding new one to the right of old one.
     * Can be used in 2 ways:
     * - [at = undefined] any tree member can split their own leaf with new peer which has the same identity key
     * - [at != undefined] tree initiator can split any tree leaf
     *
     * @param peer new Peer to add
     * @param at leaf PK of the leaf to split (only for initiator)
     * @returns {message, snapshot}. Message must be sent to all other peers, snapshot & message must be sent to new peer.
     */
    split(peer: Peer, at?: Uint8Array): {
        message: Uint8Array;
        snapshot: Uint8Array;
    };
    /**
     * Process split message created by other peers. Update stage key.
     *
     * @param message Split message made by {@link split} method
     */
    processSplit(message: Uint8Array): void;
    /**
     * Create tree snapshot containing all node PKs & peers identity PK, ephemeral PK & ephemeral signature.
     * Snapshot is used in setup message & split snapshot, therefore it also has exchange key for new peers.
     *
     * @returns
     */
    private static snapshot;
    /**
     * Decode snapshot into usable types
     */
    private static decodeSnapshot;
    /**
     * Parse split message into usable types.
     */
    private static parseSplitMessage;
}
