"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ART = void 0;
const __1 = require("..");
const err_1 = require("../err");
const tree_1 = require("./tree");
const util_1 = require("../util");
const peer_1 = require("./peer");
const node_1 = require("./node");
const leaf_1 = require("./leaf");
class ART {
    constructor(me, root, setupMessage, stage) {
        this.me = me;
        this.root = root;
        this.setupMessage = setupMessage;
        if (stage instanceof Uint8Array) {
            // fromSplitMessage
            this.stage = (0, util_1.keypair)(stage);
            this.stage = this.deriveStageKey();
        }
        else if (stage && 'sk' in stage) {
            // clone()
            this.stage = stage;
        }
        else {
            // all others
            this.stage = this.deriveStageKey();
        }
    }
    /**
     * Clone ART instance
     */
    clone() {
        return new ART(this.me.clone(), this.root.clone(), this.setupMessage.slice(), {
            sk: this.stage.sk.slice(),
            pk: this.stage.pk.slice(),
        });
    }
    /**
     * Get my leaf
     */
    get leaf() {
        const leaf = this.leafOf(this.me);
        if (!leaf) {
            throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'No my leafs in the tree');
        }
        return leaf;
    }
    /**
     * Get leaf of any peer
     * @param peer can be either Peer instance (then compared with ===), or peer identity PK / ephemeral PK
     */
    leafOf(peer) {
        return util_1.Gen.find(this.root.leafs(), (l) => l.peer === peer || (peer instanceof Uint8Array && ((0, util_1.equals)(l.peer.identity_pk, peer) || (0, util_1.equals)(l.peer.ephemeral_pk, peer))));
    }
    /**
     * Get first leaf, that is a leaf of tree initiator
     */
    get initiator() {
        let node = this.root;
        while (node.left) {
            if (node.left instanceof node_1.Node) {
                node = node.left;
            }
            else {
                break;
            }
        }
        if (node.left instanceof leaf_1.Leaf) {
            return node.left;
        }
        else {
            throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'No leftmost leaf');
        }
    }
    /**
     * Check if tree is made by initiator
     */
    get isInitiator() {
        return this.initiator.peer === this.me;
    }
    /**
     * Get list of tree Peers
     */
    get peers() {
        return util_1.Gen.collect(this.root.leafs(), (l) => l.peer);
    }
    /**
     * Get all leafs
     */
    get leafs() {
        return util_1.Gen.collect(this.root.leafs(), (l) => l);
    }
    /**
     * Derive current stage key from the tree.
     *
     * @returns stage key
     */
    deriveStageKey() {
        const root_sk = this.root.sk;
        if (!root_sk) {
            throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'Unreachable');
        }
        const ikm = new Uint8Array((this.stage ? __1.crypt.SK_LENGTH : 0) + root_sk.length + this.root.length * __1.crypt.PK_LENGTH);
        const offset = new util_1.Offset();
        if (this.stage) {
            ikm.set(this.stage.sk, offset.add(__1.crypt.SK_LENGTH));
        }
        ikm.set(root_sk, offset.add(__1.crypt.SK_LENGTH));
        for (const member of this.root.members()) {
            ikm.set(member.pk, offset.add(__1.crypt.PK_LENGTH));
        }
        offset.assert(ikm.length);
        const info = (0, util_1.concat)(...util_1.Gen.collect(this.root.leafs(), (leaf) => leaf.peer.identity_pk));
        return (this.stage = (0, util_1.keypair)(__1.crypt.hkdf(ikm, undefined, info, __1.crypt.SK_LENGTH)));
    }
    /**
     * Initiate new ART Tree with given members. First member must be instance of `Me`, that is tree initiator.
     *
     * @param peers tree members
     * @returns
     */
    static initiate(peers) {
        if (!(peers[0] instanceof peer_1.Me)) {
            throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'First peer must be initiator');
        }
        if (peers.length < 2) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'Minimal number of peers for ART is 2');
        }
        const exchange = (0, util_1.keypair)();
        const me = peers[0];
        const tree = (0, tree_1.createTree)(peers, { identity: me.identity, exchange });
        const message = ART.snapshot(tree, me, exchange);
        return new ART(me, tree, message);
    }
    /**
     * Reconstruction of a tree created by `ART.initiate()` from setup message
     *
     * @param me my peer
     * @param message setup message from initiator peer
     * @returns
     */
    static fromSetupMessage(me, message) {
        const { tree, exchange_pk, signature } = ART.decodeSnapshot(message);
        const initiatorLeaf = util_1.Gen.find(tree.leafs(), () => true);
        if (!initiatorLeaf) {
            throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'Unreachable');
        }
        if (!__1.crypt.verify(signature, message.subarray(0, -__1.crypt.SI_LENGTH), initiatorLeaf.peer.identity_pk)) {
            throw new err_1.Err(err_1.ErrCode.INVALID_SIGNATURE, 'Invalid signature of setup message');
        }
        const myLeaf = leaf_1.Leaf.me(me, {
            identity_pk: initiatorLeaf.peer.identity_pk,
            exchange_pk,
        });
        for (const member of tree.members()) {
            if (member instanceof node_1.Node) {
                if (member.left instanceof leaf_1.Leaf && (0, util_1.equals)(member.left.peer.ephemeral_pk, me.ephemeral.pk)) {
                    member.left = myLeaf;
                }
                else if (member.right instanceof leaf_1.Leaf && (0, util_1.equals)(member.right.peer.ephemeral_pk, me.ephemeral.pk)) {
                    member.right = myLeaf;
                }
            }
        }
        for (const member of tree.pathOf(me)) {
            if (member instanceof node_1.Node) {
                member.derive();
            }
        }
        return new ART(me, tree, message);
    }
    /**
     * Reconstruction of a tree created by `ART.split()` from split snapshot & message
     *
     * @param me my peer
     * @param snapshot last known snapshot of the tree
     * @param splitMessage split message
     * @returns
     */
    static fromSplitMessage(me, snapshot, splitMessage) {
        const { at, identity_pk, ephemeral_pk, ephemeral_signature, signature: messageSignature, splitter_signature, } = ART.parseSplitMessage(splitMessage);
        if (!(0, util_1.equals)(me.identity_pk, identity_pk) ||
            !(0, util_1.equals)(me.ephemeral_pk, ephemeral_pk || !(0, util_1.equals)(me.ephemeral_signature, ephemeral_signature))) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'Split message is for another peer');
        }
        const { tree, exchange_pk, signature, stage_sk } = ART.decodeSnapshot(snapshot, me);
        if (!stage_sk) {
            throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'No stage key in split snapshot');
        }
        const initiatorLeaf = util_1.Gen.find(tree.leafs(), () => true);
        if (!initiatorLeaf) {
            throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'Unreachable');
        }
        if (!__1.crypt.verify(signature, snapshot.subarray(0, -__1.crypt.SI_LENGTH), initiatorLeaf.peer.identity_pk)) {
            if (!__1.crypt.verify(signature, snapshot.subarray(0, -__1.crypt.SI_LENGTH), me.identity_pk)) {
                throw new err_1.Err(err_1.ErrCode.INVALID_SIGNATURE, 'Invalid signature of split snapshot');
            }
        }
        if (!__1.crypt.verify(messageSignature, splitMessage.subarray(0, -__1.crypt.SI_LENGTH), (0, util_1.keypair)(stage_sk).pk)) {
            throw new err_1.Err(err_1.ErrCode.INVALID_SIGNATURE, 'Invalid signature of split message');
        }
        const leaf = util_1.Gen.find(tree.leafs(), (leaf) => (0, util_1.equals)(leaf.pk, at));
        if (!leaf) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'No leaf to split at');
        }
        let init_pk;
        if (util_1.Gen.find(tree.leafs(), (l) => (0, util_1.equals)(l.peer.identity_pk, identity_pk))) {
            if (!__1.crypt.verify(splitter_signature, splitMessage.subarray(0, -2 * __1.crypt.SI_LENGTH), leaf.peer.ephemeral_pk)) {
                throw new err_1.Err(err_1.ErrCode.INVALID_SIGNATURE, 'Invalid signature of split message creator');
            }
            init_pk = leaf.peer.ephemeral_pk;
        }
        else {
            if (!__1.crypt.verify(splitter_signature, splitMessage.subarray(0, -2 * __1.crypt.SI_LENGTH), initiatorLeaf.peer.ephemeral_pk)) {
                throw new err_1.Err(err_1.ErrCode.INVALID_SIGNATURE, 'Invalid signature of split message creator');
            }
            init_pk = initiatorLeaf.peer.ephemeral_pk;
        }
        if (!tree.split(at, me, { identity_pk: init_pk, exchange_pk })) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'No such member to split at');
        }
        return new ART(me, tree, snapshot, stage_sk);
    }
    /**
     * Update my ephemeral key with new one. This is main "ratcheting" method, it updates
     * stage key. Returned message must be distributed to all other peers, preferrably in encrypted form.
     *
     * @returns update key message
     */
    updateKey(key = (0, util_1.keypair)()) {
        const updates = util_1.Gen.collect(this.root.pathOf(this.me), (node) => {
            const old = node.pk;
            if (node instanceof leaf_1.Leaf) {
                const { sk, pk } = key;
                this.me.ephemeral = { sk, pk };
                this.me.ephemeral_pk = pk;
                this.me.ephemeral_signature = __1.crypt.sign(pk, this.me.identity.sk);
                node.pk = pk;
                node.sk = sk;
                return [old, node.pk];
            }
            else {
                node.derive();
                return [old, node.pk];
            }
        }).flat();
        const message = new Uint8Array(updates.reduce((a, b) => a + b.length, 0) + __1.crypt.SI_LENGTH + __1.crypt.SI_LENGTH);
        const offset = new util_1.Offset();
        updates.forEach((pk) => message.set(pk, offset.add(__1.crypt.PK_LENGTH)));
        message.set(this.me.ephemeral_signature, offset.add(__1.crypt.SI_LENGTH));
        const signature = __1.crypt.sign(message.subarray(0, -__1.crypt.SI_LENGTH), this.stage.sk);
        message.set(signature, offset.add(__1.crypt.SI_LENGTH));
        offset.assert(message.length);
        this.deriveStageKey();
        return message;
    }
    /**
     * Process key update message from another peer. Updates stage key locally.
     */
    processKeyUpdate(message) {
        const pk_count = (message.length - __1.crypt.SI_LENGTH - __1.crypt.SI_LENGTH) / __1.crypt.PK_LENGTH;
        if (Math.round(pk_count) !== pk_count) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'Invalid key update message length');
        }
        if (!pk_count || pk_count % 2 !== 0) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'Invalid number of pks');
        }
        const offset = new util_1.Offset(0);
        const pks = (0, util_1.split)(message, offset, __1.crypt.PK_LENGTH, pk_count);
        if (pks.length < 2) {
            throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'Unreachable');
        }
        const ephemeral_signature = message.slice(offset.add(__1.crypt.SI_LENGTH), offset.value);
        const signature = message.slice(offset.add(__1.crypt.SI_LENGTH), offset.value);
        offset.assert(message.length);
        if (!__1.crypt.verify(signature, message.subarray(0, -__1.crypt.SI_LENGTH), this.stage.pk)) {
            throw new err_1.Err(err_1.ErrCode.INVALID_SIGNATURE, 'Invalid signature of key update message');
        }
        const [leafPk, peerPk] = pks;
        if (!leafPk || !peerPk) {
            throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'Not enough PKs in key update');
        }
        const leaf = util_1.Gen.find(this.root.leafs(), (l) => (0, util_1.equals)(leafPk, l.pk));
        if (!leaf) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'No such leaf');
        }
        if (!__1.crypt.verify(ephemeral_signature, peerPk, leaf.peer.identity_pk)) {
            throw new err_1.Err(err_1.ErrCode.INVALID_SIGNATURE, 'Invalid key signature in key update message');
        }
        leaf.peer.ephemeral_pk = peerPk;
        leaf.peer.ephemeral_signature = ephemeral_signature;
        for (const node of this.root.pathOf(leaf.peer)) {
            const old = pks.shift();
            const neo = pks.shift();
            if (!old || !neo || !(0, util_1.equals)(node.pk, old)) {
                throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'Invalid old PK');
            }
            node.update(neo);
        }
        if (pks.length) {
            throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'Some PKs left for update');
        }
        this.deriveStageKey();
    }
    /**
     * Split one of the leafs into 2 adding new one to the right of old one.
     * Can be used in 2 ways:
     * - [at = undefined] any tree member can split their own leaf with new peer which has the same identity key
     * - [at != undefined] tree initiator can split any tree leaf
     *
     * @param peer new Peer to add
     * @param at leaf PK of the leaf to split (only for initiator)
     * @returns {message, snapshot}. Message must be sent to all other peers, snapshot & message must be sent to new peer.
     */
    split(peer, at) {
        const exchange = (0, util_1.keypair)();
        let leaf;
        let encryption_key;
        if (util_1.Gen.find(this.root.leafs(), (l) => (0, util_1.equals)(l.peer.identity_pk, peer.identity_pk))) {
            leaf = this.leaf;
            encryption_key = this.me.identity.sk;
        }
        else {
            if (!this.isInitiator) {
                throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'You can only split your node with a peer with the same identity key');
            }
            const l = util_1.Gen.find(this.root.leafs(), (l) => (0, util_1.equals)(l.pk, at || this.leaf.pk));
            if (!l) {
                throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'No such leaf to split at');
            }
            leaf = l;
            encryption_key = (0, util_1.keyExchangeSend)(this.me.identity, peer.identity_pk, exchange, peer.ephemeral_pk);
        }
        const snapshot = ART.snapshot(this.root, this.me, exchange, { sk: this.stage.sk, encryption_key });
        if (!this.root.split(leaf.pk, peer, {
            identity: this.me.ephemeral,
            exchange,
        })) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'No such member to split at');
        }
        const pks = util_1.Gen.collect(this.root.pathOf(peer), (n) => n.pk);
        const message = new Uint8Array(__1.crypt.PK_LENGTH * 3 + __1.crypt.SI_LENGTH + pks.length * __1.crypt.PK_LENGTH + __1.crypt.SI_LENGTH * 2);
        const offset = new util_1.Offset();
        message.set(leaf.pk, offset.add(__1.crypt.PK_LENGTH));
        message.set(peer.identity_pk, offset.add(__1.crypt.PK_LENGTH));
        message.set(peer.ephemeral_pk, offset.add(__1.crypt.PK_LENGTH));
        message.set(peer.ephemeral_signature, offset.add(__1.crypt.SI_LENGTH));
        pks.forEach((pk) => message.set(pk, offset.add(__1.crypt.PK_LENGTH)));
        message.set(__1.crypt.sign(message.subarray(0, -2 * __1.crypt.SI_LENGTH), this.me.ephemeral.sk), offset.add(__1.crypt.SI_LENGTH));
        message.set(__1.crypt.sign(message.subarray(0, -__1.crypt.SI_LENGTH), this.stage.sk), offset.add(__1.crypt.SI_LENGTH));
        offset.assert(message.length);
        this.deriveStageKey();
        (0, tree_1.cleanupTree)(this.root, this.me);
        return { message, snapshot };
    }
    /**
     * Process split message created by other peers. Update stage key.
     *
     * @param message Split message made by {@link split} method
     */
    processSplit(message) {
        const { at, identity_pk, ephemeral_pk, ephemeral_signature, pks, signature, splitter_signature } = ART.parseSplitMessage(message);
        if (!__1.crypt.verify(signature, message.subarray(0, -__1.crypt.SI_LENGTH), this.stage.pk)) {
            throw new err_1.Err(err_1.ErrCode.INVALID_SIGNATURE, 'Invalid signature of split message');
        }
        const leaf = util_1.Gen.find(this.root.leafs(), (leaf) => (0, util_1.equals)(leaf.pk, at));
        if (!leaf) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'No leaf to split at');
        }
        if (!__1.crypt.verify(splitter_signature, message.subarray(0, -2 * __1.crypt.SI_LENGTH), leaf.peer.ephemeral_pk)) {
            if (!__1.crypt.verify(splitter_signature, message.subarray(0, -2 * __1.crypt.SI_LENGTH), this.initiator.peer.ephemeral_pk)) {
                throw new err_1.Err(err_1.ErrCode.INVALID_SIGNATURE, 'Invalid signature of split message creator');
            }
        }
        const peer = new peer_1.Peer(identity_pk, ephemeral_pk, ephemeral_signature);
        if (!this.root.split(at, peer, pks)) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'No such member to split at');
        }
        if (pks.length) {
            throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'Some PKs left for update');
        }
        this.deriveStageKey();
    }
    /**
     * Create tree snapshot containing all node PKs & peers identity PK, ephemeral PK & ephemeral signature.
     * Snapshot is used in setup message & split snapshot, therefore it also has exchange key for new peers.
     *
     * @returns
     */
    static snapshot(tree, me, exchange, stage) {
        const peers = util_1.Gen.collect(tree.leafs(), (l) => l.peer);
        const snapshot = new Uint8Array(peers.length * __1.crypt.PK_LENGTH + // peers' PKs
            peers.length * __1.crypt.PK_LENGTH + // peers' ephemeral PKs
            peers.length * __1.crypt.SI_LENGTH + // peers' ephemeral signatures
            (peers.length * 2 - 1) * 1 + // node & leaf kind uint8s
            (peers.length * 2 - 1) * __1.crypt.PK_LENGTH + // node & leaf PKs
            __1.crypt.PK_LENGTH + // exchange key PK
            (stage ? __1.crypt.SK_LENGTH + __1.crypt.ENCRYPTION_PREFIX_LENGTH + __1.crypt.ENCRYPTION_SUFFIX_LENGTH : 0) +
            __1.crypt.SI_LENGTH // signature
        );
        const view = new DataView(snapshot.buffer);
        const offset = new util_1.Offset();
        (0, tree_1.serializeTree)(snapshot, view, offset, tree);
        snapshot.set(exchange.pk, offset.add(__1.crypt.PK_LENGTH));
        if (stage) {
            snapshot.set(__1.crypt.encrypt(stage.sk, stage.encryption_key), offset.add(__1.crypt.SK_LENGTH + __1.crypt.ENCRYPTION_PREFIX_LENGTH + __1.crypt.ENCRYPTION_SUFFIX_LENGTH));
        }
        const signature = __1.crypt.sign(snapshot.subarray(0, -__1.crypt.SI_LENGTH), me.identity.sk);
        snapshot.set(signature, offset.add(__1.crypt.SI_LENGTH));
        offset.assert(snapshot.length);
        return snapshot;
    }
    /**
     * Decode snapshot into usable types
     */
    static decodeSnapshot(message, me) {
        const view = new DataView(message.buffer);
        const offset = new util_1.Offset(0);
        const tree = (0, tree_1.deserializeTree)(message, view, offset);
        const exchange_pk = message.slice(offset.add(__1.crypt.PK_LENGTH), offset.value);
        let stage_sk = undefined;
        if (me) {
            const peers = util_1.Gen.collect(tree.leafs(), (l) => l.peer);
            const initiator = peers[0];
            if (!initiator) {
                throw new err_1.Err(err_1.ErrCode.INCONSISTENT_STATE, 'Unreachable');
            }
            let encryption_key;
            switch (peers.findIndex((peer) => (0, util_1.equals)(peer.identity_pk, me.identity_pk))) {
                case -1:
                    encryption_key = (0, util_1.keyExchangeReceive)(me.identity, initiator.identity_pk, me.ephemeral, exchange_pk);
                    break;
                default:
                    encryption_key = me.identity.sk;
            }
            stage_sk = __1.crypt.decrypt(message.slice(offset.add(__1.crypt.SK_LENGTH + __1.crypt.ENCRYPTION_SUFFIX_LENGTH + __1.crypt.ENCRYPTION_PREFIX_LENGTH), offset.value), encryption_key);
            if (stage_sk === null) {
                throw new err_1.Err(err_1.ErrCode.INVALID_SIGNATURE, 'Failed to decrypt stage key');
            }
        }
        const signature = message.slice(offset.add(__1.crypt.SI_LENGTH), offset.value);
        offset.assert(message.length);
        return { tree, exchange_pk, signature, stage_sk };
    }
    /**
     * Parse split message into usable types.
     */
    static parseSplitMessage(message) {
        if (message.length < __1.crypt.PK_LENGTH * 3 + __1.crypt.SI_LENGTH * 3) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'Invalid split message length');
        }
        const pks_length = (message.length - __1.crypt.PK_LENGTH * 3 - __1.crypt.SI_LENGTH * 3) / __1.crypt.PK_LENGTH;
        if (Math.round(pks_length) !== pks_length) {
            throw new err_1.Err(err_1.ErrCode.INVALID_ARGUMENT, 'Invalid split message length!');
        }
        const offset = new util_1.Offset();
        const at = message.slice(offset.add(__1.crypt.PK_LENGTH), offset.value);
        const identity_pk = message.slice(offset.add(__1.crypt.PK_LENGTH), offset.value);
        const ephemeral_pk = message.slice(offset.add(__1.crypt.PK_LENGTH), offset.value);
        const ephemeral_signature = message.slice(offset.add(__1.crypt.SI_LENGTH), offset.value);
        const pks = (0, util_1.split)(message, offset, __1.crypt.PK_LENGTH, pks_length);
        const splitter_signature = message.slice(offset.add(__1.crypt.SI_LENGTH), offset.value);
        const signature = message.slice(offset.add(__1.crypt.SI_LENGTH), offset.value);
        offset.assert(message.length);
        return { at, identity_pk, ephemeral_pk, ephemeral_signature, pks, signature, splitter_signature };
    }
}
exports.ART = ART;
