"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.printTree = exports.Gen = exports.Offset = exports.toFirstThree = exports.toHex = exports.equals = exports.split = exports.concat = exports.keyExchangeReceive = exports.keyExchangeSend = exports.keypair = void 0;
const _1 = require(".");
const err_1 = require("./err");
function keypair(sk) {
    sk = sk || _1.crypt.generateSecretKey();
    const pk = _1.crypt.derivePublicKey(sk);
    return { sk, pk };
}
exports.keypair = keypair;
/**
 * X3DH send
 */
function keyExchangeSend(selfIdentity, remoteIdentity, keyExchangeKeyPair, remoteEphemeralKey) {
    const k1 = _1.crypt.getSharedSecret(selfIdentity.sk, remoteIdentity);
    const k2 = _1.crypt.getSharedSecret(selfIdentity.sk, remoteEphemeralKey);
    const k3 = _1.crypt.getSharedSecret(keyExchangeKeyPair.sk, remoteIdentity);
    const k4 = _1.crypt.getSharedSecret(keyExchangeKeyPair.sk, remoteEphemeralKey);
    return _1.crypt.hash_256(concat(k1, k2, k3, k4));
}
exports.keyExchangeSend = keyExchangeSend;
/**
 * X3DH receive
 */
function keyExchangeReceive(selfIdentity, remoteIdentity, ephemeralKey, keyExchangeKey) {
    const k1 = _1.crypt.getSharedSecret(selfIdentity.sk, remoteIdentity);
    const k2 = _1.crypt.getSharedSecret(ephemeralKey.sk, remoteIdentity);
    const k3 = _1.crypt.getSharedSecret(selfIdentity.sk, keyExchangeKey);
    const k4 = _1.crypt.getSharedSecret(ephemeralKey.sk, keyExchangeKey);
    return _1.crypt.hash_256(concat(k1, k2, k3, k4));
}
exports.keyExchangeReceive = keyExchangeReceive;
/**
 * Concat several arrays into one
 */
function concat(...arrays) {
    const result = new Uint8Array(arrays.reduce((a, b) => a + b.length, 0));
    arrays.reduce((a, b) => {
        result.set(b, a);
        return a + b.length;
    }, 0);
    return result;
}
exports.concat = concat;
/**
 * Split big array into several smaller ones of length item_length.
 * Does alloc for each subarray.
 */
function split(data, offset, item_length, item_count) {
    const result = [];
    while (offset.value < data.byteLength - item_length && (item_count === undefined || result.length < item_count)) {
        result.push(data.slice(offset.add(item_length), offset.value));
    }
    return result;
}
exports.split = split;
/**
 * Compare two arrays
 */
function equals(a, b) {
    if (a.length !== b.length) {
        return false;
    }
    for (let i = 0; i < a.length; i++) {
        if (a[i] !== b[i]) {
            return false;
        }
    }
    return true;
}
exports.equals = equals;
/**
 * Stringify Uint8Array into hex form
 *
 * @param array array to convert
 * @returns hex string of array
 */
function toHex(array) {
    const strs = [];
    array.forEach((i) => strs.push(i.toString(16).padStart(2, '0')));
    return strs.join('');
}
exports.toHex = toHex;
/**
 * Print three first numbers of Uint8Array
 *
 * @param array array to convert
 * @returns hex string of array
 */
function toFirstThree(array) {
    return array.slice(0, 3).join('.');
}
exports.toFirstThree = toFirstThree;
/**
 * Just a handy class for arrays parsing, returns old value when adding to it
 */
class Offset {
    constructor(counter = 0) {
        this.counter = counter;
    }
    get value() {
        return this.counter;
    }
    add(n) {
        const old = this.counter;
        this.counter += n;
        return old;
    }
    assert(n) {
        if (this.value !== n) {
            throw new err_1.Err(err_1.ErrCode.ENCODING_ERROR, `Wrong offset ${this.value} whereas it should be ${n}`);
        }
    }
}
exports.Offset = Offset;
exports.Gen = {
    collect(g, f) {
        const result = [];
        for (const x of g)
            result.push(f(x));
        return result;
    },
    find(g, f) {
        for (const x of g) {
            if (f(x)) {
                return x;
            }
        }
        return undefined;
    },
};
function printTree(node, level = 1, toString = toFirstThree) {
    const offset = new Array(level * 2).fill('-').join('');
    console.log(offset, node instanceof _1.Leaf ? 'Leaf' : 'Node');
    console.log(offset, 'pk', toString(node.pk));
    console.log(offset, 'sk', node.sk ? toString(node.sk) : undefined);
    if (node instanceof _1.Leaf) {
        console.log(offset, `${node.peer instanceof _1.Me ? 'Me' : 'Peer'} (${toString(node.peer.identity_pk)}, ${toString(node.peer.ephemeral_pk)})`);
    }
    else {
        console.log(offset, 'left: ');
        printTree(node.left, level + 1, toString);
        console.log(offset, 'right: ');
        printTree(node.right, level + 1, toString);
    }
}
exports.printTree = printTree;
